<?php

namespace Drupal\routing_on_exception;

use Drupal\Core\Routing\RouteMatchInterface;
use Symfony\Component\Routing\RouteCollection;

/**
 * @TODO: Make this use plugins for different tests / actions.
 */
class RoutingOnExceptionManager implements RoutingOnExceptionManagerInterface {

  public function validateOnExceptionsOption(RouteCollection $routes) {
    foreach ($routes->all() as $route_name => $route) {
      if (!$on_exception_rules = $route->getOption('on_exception')) {
        continue;
      }
      $yaml_path = "$route_name.options.on_exception";
      if (!static::isNonEmptyNumericArray($on_exception_rules)) {
        throw new RoutingOnExceptionDefinitionException(strtr(
          'Invalid route definition: If it exists, :yaml_path must be a non-empty numeric array, :type given.',
          [':yaml_path' => $yaml_path, ':type' => gettype($on_exception_rules)]
        ));
      }
      foreach ($on_exception_rules as $key => $on_exception_rule) {
        $this->validateOnExceptionRule($on_exception_rule, "${yaml_path}[$key]", $routes);
      }
    }
  }

  private function validateOnExceptionRule($on_exception_rule, string $yaml_path, RouteCollection $all_routes)
  {
    if (!static::isNonEmptyAssociativeArray($on_exception_rule)) {
      throw new RoutingOnExceptionDefinitionException(strtr(
        'Invalid route definition: :yaml_path must be a non-empty associative array, :type given.',
        [':yaml_path' => $yaml_path, ':type' => gettype($on_exception_rule)]
      ));
    }
    $this->validateOn($on_exception_rule, $yaml_path);
    $this->validateThen($on_exception_rule, $yaml_path, $all_routes);
  }

  private function validateOn(array $on_exception_rule, string $yaml_path)
  {
    if (!array_key_exists('on', $on_exception_rule)) {
      return;
    }
    if (!static::isNonEmptyNumericArray($on_exception_rule['on'])) {
      throw new RoutingOnExceptionDefinitionException(strtr(
        'Invalid route definition: :yaml_path.on must be a non-empty numeric array, :type given.',
        [':yaml_path' => $yaml_path, ':type' => gettype($on_exception_rule['on'])]
      ));
    }
    foreach ($on_exception_rule['on'] as $key => $on) {
      if (!static::isNonEmptyAssociativeArray($on)) {
        throw new RoutingOnExceptionDefinitionException(strtr(
          'Invalid route definition: :yaml_path.on[:key] must be a non-empty associative array, :type given.',
          [':yaml_path' => $yaml_path, ':key' => $key, ':type' => gettype($on)]
        ));
      }
      $valid_keys = ['name', 'message'];
      $invalid_keys = array_diff(array_keys($on), $valid_keys);
      if (count($invalid_keys) > 0) {
        throw new RoutingOnExceptionDefinitionException(strtr(
          'Invalid route definition: :yaml_path.on[:key] must contain only the keys :valid_keys, but it seems to contain :invalid_keys.',
          [
            ':yaml_path' => $yaml_path,
            ':key' => $key,
            ':valid_keys' => '"' . implode('","', $valid_keys) . '"',
            ':invalid_keys' => '"' . implode('","', $invalid_keys) . '"',
          ]
        ));
      }
      foreach ($on as $condition => $value) {
        if (!static::isValidRegularExpression($value)) {
          throw new RoutingOnExceptionDefinitionException(strtr(
            'Invalid route definition: :yaml_path.on[:key].:condition must be a valid regular expression, ":value" given.',
            [':yaml_path' => $yaml_path, ':key' => $key, ':condition' => $condition, ':value' => $value]
          ));
        }
      }
    }
  }

  private function validateThen(array $on_exception_rule, string $yaml_path, RouteCollection $all_routes)
  {
    if (!array_key_exists('then', $on_exception_rule)) {
      throw new RoutingOnExceptionDefinitionException(strtr(
        'Invalid route definition: :yaml_path must contain the key "then".',
        [':yaml_path' => $yaml_path]
      ));
    }
    if (!static::isNonEmptyNumericArray($on_exception_rule['then'])) {
      throw new RoutingOnExceptionDefinitionException(strtr(
        'Invalid route definition: :yaml_path.then must be a non-empty numeric array, :type given.',
        [':yaml_path' => $yaml_path, ':type' => gettype($on_exception_rule['then'])]
      ));
    }
    foreach ($on_exception_rule['then'] as $key => $then) {
      if (!static::isNonEmptyAssociativeArray($then)) {
        throw new RoutingOnExceptionDefinitionException(strtr(
          'Invalid route definition: :yaml_path.then[:key] must be a non-empty associative array, :type given.',
          [':yaml_path' => $yaml_path, ':key' => $key, ':type' => gettype($then)]
        ));
      }
      $valid_keys = ['log', 'redirect'];
      $invalid_keys = array_diff(array_keys($then), $valid_keys);
      if (count($invalid_keys) > 0) {
        throw new RoutingOnExceptionDefinitionException(strtr(
          'Invalid route definition: :yaml_path.then[:key] must contain only the keys :valid_keys, but it seems to contain :invalid_keys.',
          [
            ':yaml_path' => $yaml_path,
            ':key' => $key,
            ':valid_keys' => '"' . implode('","', $valid_keys) . '"',
            ':invalid_keys' => '"' . implode('","', $invalid_keys) . '"',
          ]
        ));
      }
      foreach ($then as $action => $value) {
        switch ($action) {
          case 'log':
            if (!is_string($value)) {
              throw new RoutingOnExceptionDefinitionException(strtr(
                'Invalid route definition: :yaml_path.on[:key].:action must be string, :type given.',
                [':yaml_path' => $yaml_path, ':key' => $key, ':action' => $action, ':type' => gettype($value)]
              ));
            }
            break;
          case 'redirect':
            if (!is_string($value)) {
              throw new RoutingOnExceptionDefinitionException(strtr(
                'Invalid route definition: :yaml_path.on[:key].:action must be a string, :type given.',
                [':yaml_path' => $yaml_path, ':key' => $key, ':action' => $action, ':type' => gettype($value)]
              ));
            }
            if ($all_routes->get($value) === null) {
              throw new RoutingOnExceptionDefinitionException(strtr(
                'Invalid route definition: :yaml_path.on[:key].:action must be the name of a valid route, :value given.',
                [':yaml_path' => $yaml_path, ':key' => $key, ':action' => $action, ':value' => $value]
              ));
            }
            break;
        }
      }
    }
  }

  private static function isNonEmptyAssociativeArray($value): bool
  {
    return static::isAssociativeArray($value) && count($value) > 0;
  }

  private static function isNonEmptyNumericArray($value): bool
  {
    return static::isNumericArray($value) && count($value) > 0;
  }

  private static function isAssociativeArray($value): bool
  {
    return is_array($value) && count(array_filter(array_keys($value), 'is_numeric')) === 0;
  }

  private static function isNumericArray($value): bool
  {
    return is_array($value) && count(array_filter(array_keys($value), 'is_numeric')) === count($value);
  }

  private static function isValidRegularExpression($value): bool
  {
    return is_string($value) && @preg_match('~' . strtr($value, ['~' => '\~']) . '~', '') !== false;
  }

  public function getRouteExceptionActions(RouteMatchInterface $route_match, \Throwable $exception): array {
    $route = $route_match->getRouteObject();
    if ($route === null) {
      return [];
    }
    if (!$on_exception_rules = $route->getOption('on_exception')) {
      return [];
    }

    foreach ($on_exception_rules as $on_exception_rule) {
      if ($this->ruleMatches($on_exception_rule, $exception)) {
        return $on_exception_rule['then'];
      }
    }

    return [];
  }

  private function ruleMatches($on_exception_rule, \Throwable $exception): bool
  {
    // Permit the omission of 'on' entirely. This means it matches all exceptions.
    if (!isset($on_exception_rule['on'])) {
      return true;
    }
    // For a rule to match, at least one of its 'on's must match--i.e. they are ORed together.
    foreach ($on_exception_rule['on'] as $on) {
      // For an 'on' to match, all its conditions must match--i.e. they are ANDed together.
      foreach ($on as $on_condition => $value) {
        $value_regex = '~' . strtr($value, ['~' => '\~']) . '~';
        switch ($on_condition) {
          case 'name':
            if (!preg_match($value_regex, get_class($exception))) {
              // We already know this 'on' has failed since one of its conditions failed--move to the next.
              continue 3;
            }
            break;
          case 'message':
            if (!preg_match($value_regex, $exception->getMessage())) {
              // We already know this 'on' has failed since one of its conditions failed--move to the next.
              continue 3;
            }
            break;
        }
      }
      // This 'on' has passed all its conditions. That is enough for the whole rule to match.
      return true;
    }
    return false;
  }
}
